@@ -1,6 +1,6 @@
 import {EditorState, StateCommand, EditorSelection, SelectionRange,
-        ChangeSpec, Transaction, CharCategory,
-        findClusterBreak, Text, Line, countColumn} from "@codemirror/state"
+        StateEffect, ChangeSpec, Transaction,
+        findClusterBreak, Text, Line, countColumn, CharCategory} from "@codemirror/state"
 import {EditorView, Command, Direction, KeyBinding} from "@codemirror/view"
 import {syntaxTree, IndentContext, getIndentUnit, indentUnit, indentString,
         getIndentation, matchBrackets} from "@codemirror/language"
@@ -53,6 +53,27 @@ export const cursorCharForward: Command = view => cursorByChar(view, true)
 /// Move the selection one character backward.
 export const cursorCharBackward: Command = view => cursorByChar(view, false)
 
+function byCharLogical(state: EditorState, range: SelectionRange, forward: boolean) {
+  let pos = range.head, line = state.doc.lineAt(pos)
+  if (pos == (forward ? line.to : line.from))
+    pos = forward ? Math.min(state.doc.length, line.to + 1) : Math.max(0, line.from - 1)
+  else
+    pos = line.from + findClusterBreak(line.text, pos - line.from, forward)
+  return EditorSelection.cursor(pos, forward ? -1 : 1)
+}
+
+function moveByCharLogical(target: CommandTarget, forward: boolean) {
+  return moveSel(target, range => range.empty ? byCharLogical(target.state, range, forward) : rangeEnd(range, forward))
+}
+
+/// Move the selection one character forward, in logical
+/// (non-text-direction-aware) string index order.
+export const cursorCharForwardLogical: StateCommand = target => moveByCharLogical(target, true)
+
+/// Move the selection one character backward, in logical string index
+/// order.
+export const cursorCharBackwardLogical: StateCommand = target => moveByCharLogical(target, false)
+
 function cursorByGroup(view: EditorView, forward: boolean) {
   return moveSel(view, range => range.empty ? view.moveByGroup(range, forward) : rangeEnd(range, forward))
 }
@@ -68,6 +89,25 @@ export const cursorGroupForward: Command = view => cursorByGroup(view, true)
 /// Move the selection one group backward.
 export const cursorGroupBackward: Command = view => cursorByGroup(view, false)
 
+function toGroupStart(view: EditorView, pos: number, start: string) {
+  let categorize = view.state.charCategorizer(pos)
+  let cat = categorize(start), initial = cat != CharCategory.Space
+  return (next: string) => {
+    let nextCat = categorize(next)
+    if (nextCat != CharCategory.Space) return initial && nextCat == cat
+    initial = false
+    return true
+  }
+}
+
+/// Move the cursor one group forward in the default Windows style,
+/// where it moves to the start of the next group.
+export const cursorGroupForwardWin: Command = view => {
+  return moveSel(view, range => range.empty
+    ? view.moveByChar(range, true, start => toGroupStart(view, range.head, start))
+    : rangeEnd(range, true))
+}
+
 const segmenter = typeof Intl != "undefined" && (Intl as any).Segmenter ?
   new ((Intl as any).Segmenter)(undefined, {granularity: "word"}) : null
 
@@ -197,7 +237,7 @@ function cursorByPage(view: EditorView, forward: boolean) {
       : rangeEnd(range, forward)
   })
   if (selection.eq(state.selection)) return false
-  let effect
+  let effect: StateEffect<any> | undefined
   if (page.selfScroll) {
     let startPos = view.coordsAtPos(state.selection.main.head)
     let scrollRect = view.scrollDOM.getBoundingClientRect()
@@ -268,13 +308,13 @@ export const cursorMatchingBracket: StateCommand = ({state, dispatch}) => toMatc
 /// head is currently on, if any.
 export const selectMatchingBracket: StateCommand = ({state, dispatch}) => toMatchingBracket(state, dispatch, true)
 
-function extendSel(view: EditorView, how: (range: SelectionRange) => SelectionRange): boolean {
-  let selection = updateSel(view.state.selection, range => {
+function extendSel(target: CommandTarget, how: (range: SelectionRange) => SelectionRange): boolean {
+  let selection = updateSel(target.state.selection, range => {
     let head = how(range)
     return EditorSelection.range(range.anchor, head.head, head.goalColumn, head.bidiLevel || undefined)
   })
-  if (selection.eq(view.state.selection)) return false
-  view.dispatch(setSel(view.state, selection))
+  if (selection.eq(target.state.selection)) return false
+  target.dispatch(setSel(target.state, selection))
   return true
 }
 
@@ -293,6 +333,15 @@ export const selectCharForward: Command = view => selectByChar(view, true)
 /// Move the selection head one character backward.
 export const selectCharBackward: Command = view => selectByChar(view, false)
 
+/// Move the selection head one character forward by logical
+/// (non-direction aware) string index order.
+export const selectCharForwardLogical: StateCommand =
+  target => extendSel(target, range => byCharLogical(target.state, range, true))
+/// Move the selection head one character backward by logical string
+/// index order.
+export const selectCharBackwardLogical: StateCommand =
+  target => extendSel(target, range => byCharLogical(target.state, range, false))
+
 function selectByGroup(view: EditorView, forward: boolean) {
   return extendSel(view, range => view.moveByGroup(range, forward))
 }
@@ -308,6 +357,12 @@ export const selectGroupForward: Command = view => selectByGroup(view, true)
 /// Move the selection head one group backward.
 export const selectGroupBackward: Command = view => selectByGroup(view, false)
 
+/// Move the selection head one group forward in the default Windows
+/// style, skipping to the start of the next group.
+export const selectGroupForwardWin: Command = view => {
+  return extendSel(view, range => view.moveByChar(range, true, start => toGroupStart(view, range.head, start)))
+}
+
 function selectBySubword(view: EditorView, forward: boolean) {
   return extendSel(view, range => moveBySubword(view, range, forward))
 }
@@ -401,16 +456,21 @@ export const selectLine: StateCommand = ({state, dispatch}) => {
 /// syntax tree.
 export const selectParentSyntax: StateCommand = ({state, dispatch}) => {
   let selection = updateSel(state.selection, range => {
-    let stack = syntaxTree(state).resolveStack(range.from, 1)
+    let tree = syntaxTree(state), stack = tree.resolveStack(range.from, 1)
+    if (range.empty) {
+      let stackBefore = tree.resolveStack(range.from, -1)
+      if (stackBefore.node.from >= stack.node.from && stackBefore.node.to <= stack.node.to) stack = stackBefore
+    }
     for (let cur: typeof stack | null = stack; cur; cur = cur.next) {
       let {node} = cur
       if (((node.from < range.from && node.to >= range.to) ||
            (node.to > range.to && node.from <= range.from)) &&
-          node.parent?.parent)
+           cur.next)
         return EditorSelection.range(node.to, node.from)
     }
     return range
   })
+  if (selection.eq(state.selection)) return false
   dispatch(setSel(state, selection))
   return true
 }
@@ -913,7 +973,7 @@ export const emacsStyleKeymap: readonly KeyBinding[] = [
 ///  - End: [`cursorLineBoundaryForward`](#commands.cursorLineBoundaryForward) ([`selectLineBoundaryForward`](#commands.selectLineBoundaryForward) with Shift)
 ///  - Ctrl-Home (Cmd-Home on macOS): [`cursorDocStart`](#commands.cursorDocStart) ([`selectDocStart`](#commands.selectDocStart) with Shift)
 ///  - Ctrl-End (Cmd-Home on macOS): [`cursorDocEnd`](#commands.cursorDocEnd) ([`selectDocEnd`](#commands.selectDocEnd) with Shift)
-///  - Enter: [`insertNewlineAndIndent`](#commands.insertNewlineAndIndent)
+///  - Enter and Shift-Enter: [`insertNewlineAndIndent`](#commands.insertNewlineAndIndent)
 ///  - Ctrl-a (Cmd-a on macOS): [`selectAll`](#commands.selectAll)
 ///  - Backspace: [`deleteCharBackward`](#commands.deleteCharBackward)
 ///  - Delete: [`deleteCharForward`](#commands.deleteCharForward)
@@ -947,7 +1007,7 @@ export const standardKeymap: readonly KeyBinding[] = ([
   {key: "End", run: cursorLineBoundaryForward, shift: selectLineBoundaryForward, preventDefault: true},
   {key: "Mod-End", run: cursorDocEnd, shift: selectDocEnd},
 
-  {key: "Enter", run: insertNewlineAndIndent},
+  {key: "Enter", run: insertNewlineAndIndent, shift: insertNewlineAndIndent},
 
   {key: "Mod-a", run: selectAll},
 
